<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <!-- https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy/script-src -->
    <meta http-equiv="Content-Security-Policy" content="script-src 'self';" />
    <link rel="stylesheet" href="/svelte-bundle.css" />
    <script defer src="/svelte-bundle.js"></script>
  </head>
  <body>
    <!--  -->
  </body>
</html>
